/*  HackUtil is a set of functions that make import detours, code patching, and logging easier.
    Copyright (C) 2010  Adam Heinermann

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "HackUtil.h"

struct memPatch
{
  void* location;
  char  patch[128];
  u32   patchSize;
};

static FILE* globalLog;
static std::vector<memPatch> changes;

///////////////////////////////   LOGGER    //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
void logfast(char* file, const char* format, ...)
{
  if (file == NULL)
  {
    logError("No file name specified for %s", "logfast()");
    return;
  }

  if (format == NULL)
  {
    logError("No formatting specified for %s", "logfast()");
    return;
  }

  FILE* logFile = _fsopen(file, "a+", _SH_DENYNO);

  SYSTEMTIME time;
  GetLocalTime(&time);
  fprintf_s(logFile, "[%02d:%02d:%02d:%03d] ", time.wHour, time.wMinute, time.wSecond, time.wMilliseconds);

  va_list ap;
  va_start(ap, format);
  vfprintf_s(logFile, format, ap);
  va_end(ap);

  fprintf_s(logFile, "\n");
  fclose(logFile);
  return;
}

void logError(const char* format, ...)
{
  if (format == NULL)
  {
    logError("No formatting specified for %s", "logError()");
    return;
  }

  char hackbuffer[BUFFER_SIZE];
  va_list ap;
  va_start(ap, format);
  vsnprintf_s(hackbuffer, BUFFER_SIZE, BUFFER_SIZE, format, ap);
  va_end(ap);
  logfast("Errors.log", "%s", hackbuffer);
  return;
}

void logBegin(char* file)
{
  if (file == NULL)
  {
    logError("No file name specified for %s", "logBegin()");
    return;
  }

  if (globalLog != NULL)
  {
    fprintf_s(globalLog, "\n");
    fclose(globalLog);
  }
  globalLog = _fsopen(file, "a+", _SH_DENYNO);
  return;
}

void logBegin()
{
  if (globalLog == NULL)
    return;

  fprintf_s(globalLog, "\n");
  return;
}

void logTime()
{
  if (globalLog == NULL)
    return;

  SYSTEMTIME time;
  GetLocalTime(&time);
  fprintf_s(globalLog, "[%02d:%02d:%02d:%03d] ", time.wHour, time.wMinute, time.wSecond, time.wMilliseconds);
  return;
}

void logBytes(u8 *data, u32 size)
{
  logBytes(globalLog, data, size);
  return;
}

void logBytes(FILE* file, u8* data, u32 size)
{
  if (file == NULL)
  {
    logError("Bad file handle for logBytes().");
    return;
  }

  if (data == NULL)
  {
    logError("No data specified for logBytes().");
    return;
  }

  char hackbuffer[BUFFER_SIZE];
  memset(hackbuffer, 0, BUFFER_SIZE);
  for (u32 i = 0; i < size; i++)
  {
    u8 wbyte;
    __try
    {
      wbyte = data[i];
    }
    __except(1)
    {
      wbyte = 0xFF;
      logError("Read from invalid memory address 0x%08X[%d] in logBytes().", data, size);
    }
    char temp[5];
    sprintf_s(temp, 4, " %02X", wbyte);
    strcat_s(hackbuffer, BUFFER_SIZE, temp);
  }
  fprintf_s(globalLog, "%s", &hackbuffer[1]);
  return;
}

void logBytes(char* file, u8* data, u32 size)
{
  if (file == NULL)
  {
    logError("No file name specified for %s", "logBytes()");
    return;
  }

  FILE* logFile = _fsopen(file, "a+", _SH_DENYNO);
  logBytes(logFile, data, size);
  fclose(logFile);
  return;
}

void logAdd(const char *format, ...)
{
  if (globalLog == NULL)
    return;

  if (format == NULL)
  {
    logError("No formatting specified for %s", "logAdd()");
    return;
  }

  va_list ap;
  va_start(ap, format);
  vfprintf_s(globalLog, format, ap);
  va_end(ap);
  return;
}

void logEnd()
{
  if (globalLog == NULL)
    return;

  fprintf_s(globalLog, "\n");
  fclose(globalLog);
  return;
}
///////////////////////////////     FILE    //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
void fadd(char *file, const char *format, ...)
{
  FILE* logFile = _fsopen(file, "a+", _SH_DENYNO);

  va_list ap;
  va_start(ap, format);
  vfprintf_s(logFile, format, ap);
  va_end(ap);

  fprintf_s(logFile, "\n");
  fclose(logFile);
  return;
}

///////////////////////////////   IMPORTS   //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

IMAGE_IMPORT_DESCRIPTOR* _GetImportDescriptor(HMODULE module)
{
  IMAGE_DOS_HEADER *mzhead = (IMAGE_DOS_HEADER*)module;
  if (mzhead->e_magic != IMAGE_DOS_SIGNATURE)
  {
    logError("%sMZ header.", "Could not locate ");
    return NULL;
  }
  IMAGE_NT_HEADERS32 *pehead = (IMAGE_NT_HEADERS32*)(mzhead->e_lfanew + (u32)module);
  if (pehead->Signature != IMAGE_NT_SIGNATURE)
  {
    logError("%sPE header.", "Could not locate ");
    return NULL;
  }
  IMAGE_IMPORT_DESCRIPTOR *imports = (IMAGE_IMPORT_DESCRIPTOR*)(pehead->OptionalHeader.DataDirectory[1].VirtualAddress + (u32)module);
  if (imports == NULL)
  {
    logError("%simports section.", "Could not locate ");
    return NULL;
  }
  return imports;
}

IMAGE_THUNK_DATA32* _GetImportsList(char* sourceModule, char* importModule)
{
  char *tempModuleName = NULL;
  if (sourceModule != NULL)
    tempModuleName = sourceModule;

  HMODULE tempModule;
  tempModule = GetModuleHandleA(tempModuleName);

  if (tempModule == NULL)
  {
    logError("Could not obtain module handle of \"%s\".", tempModuleName);
    return NULL;
  }

  IMAGE_IMPORT_DESCRIPTOR *imports = _GetImportDescriptor(tempModule);
  if (imports == NULL)
    return NULL;
  
  for (u32 i = 0; imports[i].Name != 0; i++)
  {
    if (lstrcmpiA((char*)(imports[i].Name + (u32)tempModule), importModule) == 0)
      return (IMAGE_THUNK_DATA32*)(imports[i].OriginalFirstThunk + (u32)tempModule);
  }
  return NULL;
}

IMAGE_THUNK_DATA32* _GetImportsList(char* importModule)
{
  return _GetImportsList(NULL, importModule);
}

DWORD* _GetFunctionsList(char* sourceModule, char* importModule)
{
  char *tempModuleName = NULL;
  if (sourceModule != NULL)
    tempModuleName = sourceModule;

  HMODULE tempModule = GetModuleHandleA(tempModuleName);
  if (tempModule == NULL)
  {
    logError("Could not obtain module handle of \"%s\".", tempModuleName);
    return NULL;
  }

  IMAGE_IMPORT_DESCRIPTOR *imports = _GetImportDescriptor(tempModule);
  if (imports == NULL)
    return NULL;

  for (u32 i = 0; imports[i].Name != 0; i++)
  {
    if (lstrcmpiA((char*)(imports[i].Name + (u32)tempModule), importModule) == 0)
      return (DWORD*)(imports[i].FirstThunk + (u32)tempModule);
  }
  return NULL;
}

DWORD* _GetFunctionsList(char* importModule)
{
  return _GetFunctionsList(NULL, importModule);
}

bool PatchImport(char* sourceModule, char* importModule, LPCSTR function, void* patchFunction)
{
  if (function == NULL)
  {
    logError("Did not specify a function%s", " for PatchImport().");
    return false;
  }

  char *tempModuleName = NULL;
  if (sourceModule != NULL)
    tempModuleName = sourceModule;

  HMODULE tempModule = GetModuleHandleA(tempModuleName);
  if (tempModule == NULL)
  {
    logError("Could not obtain module handle of \"%s\"%s", tempModuleName, " for PatchImport().");
    return false;
  }

  IMAGE_THUNK_DATA32* importOrigin = _GetImportsList(sourceModule, importModule);
  if (importOrigin == NULL)
  {
    logError("%sImports Section for \"%s\" in %s%s", "Could not locate ", importModule, sourceModule, " for PatchImport().");
    return false;
  }

  DWORD* importFunction = _GetFunctionsList(sourceModule, importModule);
  if (importOrigin == NULL)
  {
    logError("%sFunctions list for \"%s\" in %s%s", "Could not locate ", importModule, sourceModule, " for PatchImport().");
    return false;
  }

  for (u32 i = 0; importOrigin[i].u1.Ordinal != 0; i++)
  {
    if (IMAGE_SNAP_BY_ORDINAL32(importOrigin[i].u1.Ordinal) && (DWORD)function < 0xFFFF)
    {
      if (IMAGE_ORDINAL32(importOrigin[i].u1.Ordinal) == IMAGE_ORDINAL32((DWORD)function))
      {
        _WriteMem("", &importFunction[i], &patchFunction, 4);
        return true;
      }
    }
    else
    {
      if (lstrcmpiA(function, (char*)&(((PIMAGE_IMPORT_BY_NAME)importOrigin[i].u1.AddressOfData)->Name)) == 0)
      {
        _WriteMem("", &importFunction[i], &patchFunction, 4);
        return true;
      }
    }
  }
  
  if ((DWORD)function < 0xFFFF)
    logError("%simport of ordinal %d (0x%X) in %s for PatchImport().", "Could not locate ", (DWORD)function, (DWORD)function, tempModuleName);
  else
    logError("%simport of function \"%s\" in %s for PatchImport().", "Could not locate ", function, tempModuleName);

  return false;
}

bool PatchImport(char* importModule, LPCSTR function, void* patchFunction)
{
  return PatchImport(NULL, importModule, function, patchFunction);
}

FARPROC _GetImport(const char *logline, char* importModule, LPCSTR function)
{
  HMODULE module = GetModuleHandleA(importModule);
  if (module == NULL)
  {
    logError("%smodule \"%s\"", "Could not find ", importModule, logline);
    return NULL;
  }

  FARPROC rval = GetProcAddress(module, function);
  if (rval == NULL)
  {
    if ((DWORD)function < 0xFFFF)
      logError("%sordinal %d (0x%X) in module %s (%s).", "Could not find ", (DWORD)function, (DWORD)function, importModule, logline);
    else
      logError("%sfunction \"%s\" in module %s (%s).", "Could not find ", function, importModule, logline);
    return NULL;
  }

  return rval;
}

///////////////////////////////   MEMORY   //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

void _WriteNops(const char *logline, void* dest, u32 size)
{
  memPatch tempPatch;
  tempPatch.location = dest;
  tempPatch.patchSize = size;
  memcpy_s(tempPatch.patch, 128, dest, size);
  changes.push_back(tempPatch);

  DWORD lastAccess = 0;
  VirtualProtect(dest, size, PAGE_EXECUTE_READWRITE, &lastAccess);
  __try
  {
    memset(dest, 0x90, size);
  } 
  __except(1)
  {
    logError("Wrote to invalid memory address 0x%08X[%d] (%s).", (u32)dest, size, logline);
  }
  VirtualProtect(dest, size, lastAccess, &lastAccess);
  return;
}

void _WriteNops(const char *logline, u32 dest, u32 size)
{
  _WriteNops(logline, (void*)dest, size);
  return;
}

void _WriteMem(const char *logline, void* dest, void* source, u32 size)
{
  memPatch tempPatch;
  tempPatch.location = dest;
  tempPatch.patchSize = size;
  memcpy_s(tempPatch.patch, 128, dest, size);
  changes.push_back(tempPatch);

  _WriteMemRaw(logline, dest, source, size);
  return;
}

void _WriteMemRaw(const char *logline, void* dest, void* source, u32 size)
{
  DWORD lastAccess = 0;
  VirtualProtect(dest, size, PAGE_EXECUTE_READWRITE, &lastAccess);
  __try
  {
    memcpy_s(dest, size, source, size);
  } 
  __except(1)
  {
    logError("Wrote to invalid memory address 0x%08X[%d] (%s).", (u32)dest, size, logline);
  }
  VirtualProtect(dest, size, lastAccess, &lastAccess);
  return;
}

void _WriteMem(const char *logline, u32 dest, void* source, u32 size)
{
  _WriteMem(logline, (void*)dest, source, size);
  return;
}

void _WriteMemRaw(const char *logline, u32 dest, void* source, u32 size)
{
  _WriteMemRaw(logline, (void*)dest, source, size);
  return;
}

void _JmpPatch(const char *logline, void* dest, void* patch)
{
  u8 temp[5];
  temp[0] = 0xE9;
  *(int*)&temp[1] = (int)dest - (int)patch - 5;
  _WriteMem(logline, dest, temp, 5);
  return;
}

void _JmpPatch(const char *logline, u32 dest, void* patch)
{
  _JmpPatch(logline, (void*)dest, patch);
  return;
}

void _CallPatch(const char *logline, void* dest, void* patch)
{
  u8 temp[5];
  temp[0] = 0xE8;
  *(int*)&temp[1] = (int)dest - (int)patch - 5;
  _WriteMem(logline, dest, temp, 5);
  return;
}

void _CallPatch(const char *logline, u32 dest, void* patch)
{
  _CallPatch(logline, (void*)dest, patch);
  return;
}

void _Revert(const char *logline)
{
  for each (memPatch i in changes)
    _WriteMemRaw(logline, i.location, i.patch, i.patchSize);
  changes.clear();
  return;
}
